#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/time.h>

#include <machine/sysarch.h>
#include <machine/cpufunc.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <libgen.h>
#include <unistd.h>

#define	RTC_ADDR	0x70
#define	RTC_DATA	0x71

#define	RTC_SEC		0x00
#define RTC_MIN         0x02
#define RTC_HRS         0x04
#define RTC_DAY         0x07
#define RTC_MONTH       0x08
#define RTC_YEAR        0x09
#define	RTC_CENTURY	0x32

#define RTC_STATUSB     0x0b

u_char const bin2bcd_data[] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99
};
#define bin2bcd(bin)    (bin2bcd_data[bin])

static char *progname;

static int
bcd2bin(uint8_t x)
{
	uint8_t lower, upper;

	upper = x >> 4;
	lower = x & 0xf;
	if (upper > 9 || lower > 9)
		fprintf(stderr, "Invalid BCD value %02X\n", x);

	return (upper * 10 + lower);
}

static int
rtcin(int reg)
{
	outb(RTC_ADDR, reg);
	return (inb(RTC_DATA));
}

static void
rtcout(int reg, int val)
{
	outb(RTC_ADDR, reg);
	outb(RTC_DATA, val);
}

static void
usage(int error)
{
	fprintf(stderr,
"Usage: %s [-hd][-s <seconds>][-c <century>]\n"
"       -h: print this help message\n"
"       -d: dump the RTC date and time\n"
"       -s: set the RTC to the UNIX time <seconds>\n"
"       -c: set the RTC century to <century>\n\n"
"The UNIX time can be obtained by using the 'date' command. For example, to\n"
"convert 23:59:30 Dec 31 1999 to UNIX time use the following command:\n\n"
"$ date -j +%%s 199912312359.45\n\n", progname);
}

int
main(int argc, char *argv[])
{
	struct i386_ioperm_args ioperm;
	int century, error, reg_b;
	int sec, min, hour, mon, dom, year;
	int c, tmp, dump, set_date_time, set_century;
	time_t seconds;
	struct tm tm;

	progname = basename(argv[0]);

	dump = 0;
	set_century = 0;
	set_date_time = 0;
	while ((c = getopt(argc, argv, "dhc:s:")) != -1) {
		switch (c) {
		case 'd':
			dump = 1;
			break;
		case 'c':
			set_century = 1;
			century = strtoul(optarg, NULL, 0);
			break;
		case 's':
			set_date_time = 1;
			seconds = strtoul(optarg, NULL, 0);
			break;
		case 'h':
			usage(0);
		default:
			usage(1);
		}
	}
	argc -= optind;
	argv += optind;

	ioperm.start = 0x70;
	ioperm.length = 2;
	ioperm.enable = 1;

	error = sysarch(I386_SET_IOPERM, &ioperm);
	if (error) {
		perror("i386_set_ioperm");
		exit(1);
	}

	/* sysctl machdep.disable_rtc_set=1 */
	tmp = 1;
	error = sysctlbyname("machdep.disable_rtc_set", NULL, NULL,
	    &tmp, sizeof(int));
	if (error) {
		fprintf(stderr, "error %d setting machdep.disable_rtc_set\n",
		    error);
		exit(1);
	}

	reg_b = rtcin(RTC_STATUSB);
	if ((reg_b & 0x6) != 0x2) {
		fprintf(stderr, "RTC must be in 24-hour BCD format: %02X\n",
		    reg_b);
		exit(1);
	}

	if (set_date_time) {
		gmtime_r(&seconds, &tm);
		printf("Setting RTC time %02d:%02d:%02d %02d/%02d/%04d\n",
		    tm.tm_hour, tm.tm_min, tm.tm_sec, tm.tm_mon + 1,
		    tm.tm_mday, tm.tm_year + 1900);

		rtcout(RTC_STATUSB, reg_b | 0x80);	/* halt RTC */

		/*
		 * tm_sec range is from [0,60] but RTC only allows [0,59]
		 * so drop the occasional leap second.
		 */
		if (tm.tm_sec == 60)
			tm.tm_sec = 59;

		rtcout(RTC_SEC, bin2bcd(tm.tm_sec));
		rtcout(RTC_MIN, bin2bcd(tm.tm_min));
		rtcout(RTC_HRS, bin2bcd(tm.tm_hour));
		rtcout(RTC_DAY, bin2bcd(tm.tm_mday));
		rtcout(RTC_MONTH, bin2bcd(tm.tm_mon + 1));
		rtcout(RTC_YEAR, bin2bcd(tm.tm_year % 100));
		rtcout(RTC_CENTURY, bin2bcd((1900 + tm.tm_year) / 100));

		rtcout(RTC_STATUSB, reg_b);		/* resume RTC */
	}

	if (set_century) {
		printf("Setting RTC century to %d\n", century);
		rtcout(RTC_CENTURY, bin2bcd(century));
	}

	if (dump) {
		century = bcd2bin(rtcin(RTC_CENTURY));
		sec = bcd2bin(rtcin(RTC_SEC));
		min = bcd2bin(rtcin(RTC_MIN));
		hour = bcd2bin(rtcin(RTC_HRS));
		dom = bcd2bin(rtcin(RTC_DAY));
		mon = bcd2bin(rtcin(RTC_MONTH));
		year = bcd2bin(rtcin(RTC_YEAR));
		printf("RTC time %02d:%02d:%02d %02d/%02d/%04d\n",
		    hour, min, sec, mon, dom, century * 100 + year);
	}
}
